%% Solution for Exercise 9: Direct Multiple Shooting and Real-Time Iterations with YALMIP
%    TEMPO Summer School on Numerical Optimal Control and Embedded Optimization 
%    University of Freiburg, July 27 - August 7, 2015 
%    Rien Quirynen, Dimitris Kouzoupis, Joachim Ferreau and Moritz Diehl 

%% SQP and Gauss-Newton for direct multiple shooting
clc;
close all;
clear all;
yalmip('clear')

% Define parameters:
nu = 1;
nx = 2;
h = 0.1;
N = 15;
Tp = N*h;
input.Ts = h;
input.nSteps = 2;

Q = 0.5*eye(nx);
R = eye(nu);
QN = 16.5926*eye(nx);
QN(2,1) = 11.5926;
QN(1,2) = 11.5926;

% Initial state values
x0 = [-0.683; -0.864];

% Initialization of state and control trajectory:
X = zeros(nx, N+1);
U = zeros(nu, N);


figure(1);
subplot(2,2,1);
plot(h*[0:N], X(1,:), '--g');
xlabel('time(s)')
ylabel('x1')
title('State 1')

subplot(2,2,2);
plot(h*[0:N], X(2,:), '--g');
xlabel('time(s)')
ylabel('x2')
title('State 2')

subplot(2,2,3:4);
stairs(h*[0:N-1], U, '--b');
xlabel('time(s)')
ylabel('u')
title('Control')

u = sdpvar(nu*ones(1,N),ones(1,N));
x = sdpvar(nx*ones(1,N+1),ones(1,N+1));

% SET-UP OBJECTIVE:
objective = 0;
for k = 1:N
    objective = objective + h/2*(x{k}'*Q*x{k} + u{k}'*R*u{k});
end
objective = objective + 1/2*x{N+1}'*QN*x{N+1};
    
iter = 0;
DX = 1; DU = 1;
while max(DX,DU) > 1e-6
    
    % SET-UP CONSTRAINTS:
    constraints = [];
    for k = 1:N
        input.x = X(:,k);
        input.u = U(:,k);
        output = RK4_integrator(@ode, input);
        constraints = [constraints, output.value-x{k+1} + output.sensX*(x{k}-X(:,k)) + output.sensU*(u{k}-U(:,k)) == 0];
        constraints = [constraints, -2.0 <= u{k} <= 2.0];
    end
    constraints = [constraints, x{1} == x0];
    
    % SOLVE THE CONVEX PROBLEM:
    opts = sdpsettings('verbose',0,'solver','quadprog');
    solvesdp(constraints, objective, opts);
    
    % GET THE CURRENT VALUES:
    deltaX = X;
    deltaU = U;
    for k = 1:N+1
        deltaX(:,k) = abs(X(:,k) - double(x{k}));
        X(:,k) = double(x{k});
    end
    for k = 1:N
        deltaU(:,k) = abs(U(:,k) - double(u{k}));
        U(:,k) = double(u{k});
    end
    DX = max(max(deltaX));
    DU = max(max(deltaU));
    disp(['correction at iteration ' num2str(iter) ': DX = ' num2str(DX) '; DU = ' num2str(DU)]);
    
    iter = iter+1;
    
    figure(1);
    subplot(2,2,1);
    plot(h*[0:N], X(1,:), '--g');
    xlabel('time(s)')
    ylabel('x1')
    title('State 1')
    
    subplot(2,2,2);
    plot(h*[0:N], X(2,:), '--g');
    xlabel('time(s)')
    ylabel('x2')
    title('State 2')
    
    subplot(2,2,3:4);
    stairs(h*[0:N-1], U, '--b');
    xlabel('time(s)')
    ylabel('u')
    title('Control')
end


%% Closed-loop Nonlinear MPC simulation
clc;
close all;
clear all;
yalmip('clear')

h = 0.1;
input.Ts = h;
input.nSteps = 2;
nu = 1;
nx = 2;
N = 15;
Tp = N*h;

% RTI scheme:
MAXITs = 1;
% ideal NMPC scheme:
% MAXITs = 10;

Q = 0.5*eye(nx);
R = eye(nu);
QN = 16.5926*eye(nx);
QN(2,1) = 11.5926;
QN(1,2) = 11.5926;
Nt = 50;

x0 = [-0.683; -0.864];

X = zeros(nx, N+1);
U = zeros(nu, N);

u = sdpvar(nu*ones(1,N),ones(1,N));
x = sdpvar(nx*ones(1,N+1),ones(1,N+1));

% SET-UP OBJECTIVE:
objective = 0;
for k = 1:N
    objective = objective + h/2*(x{k}'*Q*x{k} + u{k}'*R*u{k});
end
objective = objective + 1/2*x{N+1}'*QN*x{N+1};
sys_states = [];
sys_controls = [];
for j = 0:Nt-1
    disp(['----------------- time step ' num2str(j+1) ' -----------------']);
    tic
    
    disp(['current time: ' num2str(h*j)])
    disp(['current state system: ' num2str(x0(1)) '; ' num2str(x0(2))])
    
    DX = 1;
    DU = 1;
    i = 0;
    while max(DX, DU) > 1e-5 && i < MAXITs
        
        % SET-UP CONSTRAINTS:
        constraints = [];
        for k = 1:N
            input.x = X(:,k);
            input.u = U(:,k);
            states = RK4_integrator(@ode, input);
            constraints = [constraints, states.value-x{k+1} + states.sensX*(x{k}-X(:,k)) + states.sensU*(u{k}-U(:,k)) == 0];
            constraints = [constraints, -2.0 <= u{k} <= 2.0];
        end
        constraints = [constraints, x{1} == x0];
        constraints = [constraints, x{N+1}'*QN*x{N+1} <= 0.7];
        
        % SOLVE THE CONVEX PROBLEM:
        opts = sdpsettings('verbose', 0, 'solver','SDPT3');
        solvesdp(constraints, objective, opts);
        
        % GET THE CURRENT VALUES:
        deltaX = X;
        deltaU = U;
        for k = 1:N+1
            deltaX(:,k) = abs(X(:,k) - double(x{k}));
            X(:,k) = double(x{k});
        end
        for k = 1:N
            deltaU(:,k) = abs(U(:,k) - double(u{k}));
            U(:,k) = double(u{k});
        end
        DX = max(max(deltaX));
        DU = max(max(deltaU));
        disp(['correction: ' num2str(DX) '; ' num2str(DU)]);
        
        i = i+1;
    end
    input.x = x0;
    input.u = U(:,1);
    states = RK4_integrator(@ode, input);
    
    figure(2);
    subplot(2,2,1);
    hold on;
    plot(h*[j:1+j], [x0(1) states.value(1,1)], '--g');
    xlabel('t')
    ylabel('x1')
    title('State 1')
    
    subplot(2,2,2);
    hold on;
    plot(h*[j:1+j], [x0(2) states.value(2,1)], '--g');
    xlabel('t')
    ylabel('x2')
    title('State 2')
    
    subplot(2,2,3:4);
    hold on;
    if j > 0
        stairs(h*[j-1:j], [uprev U(:,1)], '--b');
    end
    xlabel('t')
    ylabel('u')
    title('Control')
    
    x0 = states.value;
    uprev = U(:,1);
    
    sys_states = [sys_states x0];
    sys_controls = [sys_controls uprev];
    
    % SHIFT:
    for k = 1:N
        X(:,k) = X(:,k+1);
    end
    for k = 1:N-1
        U(:,k) = U(:,k+1);
    end
    input.x = X(:,N);
    input.u = U(:,N);
    states = RK4_integrator(@ode, input);
    X(:,N+1) = states.value;
    
    toc
    disp(['-----------------------------------------------']);
end

sys_states2 = sys_states;
sys_controls2 = sys_controls;

load NMPC_sol.mat sys_states sys_controls

figure;
subplot(2,2,1);
plot(h*[0:Nt-1], sys_states(1,:), '--rx'); hold on
plot(h*[0:Nt-1], sys_states2(1,:), '--bo');
xlabel('time(s)')
ylabel('x1')
title('State 1')
legend('ideal NMPC', 'RTI scheme')

subplot(2,2,2);
plot(h*[0:Nt-1], sys_states(2,:), '--rx'); hold on
plot(h*[0:Nt-1], sys_states2(2,:), '--bo');
xlabel('time(s)')
ylabel('x2')
title('State 2')
legend('ideal NMPC', 'RTI scheme')

subplot(2,2,3:4);
stairs(h*[0:Nt-1], sys_controls, '--rx'); hold on
stairs(h*[0:Nt-1], sys_controls2, '--bo');
xlabel('time(s)')
ylabel('u')
title('Control')
legend('ideal NMPC', 'RTI scheme')


